@@ -62,14 +62,16 @@ module Codec.Archive.LibZip
     , Entry
     , ZipStat(..)
     -- * Archive operations
-    , withArchive, getZip
+    , withArchive, withEncryptedArchive, getZip
     , numFiles, fileName, nameLocate, fileNames
     , fileSize, fileSizeIx
     , fileStat, fileStatIx
     , deleteFile, deleteFileIx
     , renameFile, renameFileIx
-    , addFile, addDirectory
+    , addFile, addFileWithFlags
+    , addDirectory, addDirectoryWithFlags
     , replaceFile, replaceFileIx
+    , setFileCompression, setFileCompressionIx
     , sourceBuffer, sourceFile, sourceZip
     , PureSource(..), sourcePure
     , getComment, setComment, removeComment
@@ -105,7 +107,7 @@ import Control.Monad (when)
 import Control.Monad.State.Strict
     (StateT(..), MonadState(..), MonadTrans(..), lift, liftM)
 import Foreign.C.Error (Errno(..), eINVAL)
-import Foreign.C.String (withCString, withCStringLen, peekCString)
+import Foreign.C.String (withCString, peekCString)
 import Foreign.C.Types (CInt, CULLong)
 import Foreign.Marshal.Alloc (alloca)
 import Foreign.Marshal.Array (allocaArray, peekArray, withArrayLen, pokeArray)
@@ -113,6 +115,8 @@ import Foreign.Marshal.Utils (with)
 import Foreign.Ptr (Ptr, nullPtr, castPtr)
 import Foreign.Storable (Storable, peek, poke, pokeElemOff, sizeOf)
 import qualified Control.Exception as E
+import qualified Data.ByteString as BS
+import qualified Data.ByteString.UTF8 as UTF8
 
 --
 -- Types
@@ -145,21 +149,51 @@ withArchive flags path action =
   c'zip_open path' (combine flags) errp >>= \z ->
   if z == nullPtr
     then peek errp >>= E.throwIO. errFromCInt
-    else do
+    else withOpenArchive z action
+
+
+-- | Top-level wrapper for operations with an open encrypted archive.
+-- 'withEncryptedArchive' opens and closes the file automatically.
+-- On error it throws 'ZipError'.
+withEncryptedArchive :: [OpenFlag]   -- ^ Checks for consistency or existence.
+                     -> String       -- ^ Encryption password.
+                     -> FilePath     -- ^ Filename of the zip archive.
+                     -> Archive a    -- ^ Action to don with the archive.
+                     -> IO a
+withEncryptedArchive flags password path action =
+    withCString password $ \password' ->
+    withCString path $ \path' ->
+    alloca $ \errp ->
+    c'zip_open path' (combine flags) errp >>= \z ->
+    if z == nullPtr
+       then peek errp >>= E.throwIO. errFromCInt
+       else do
+         r <- c'zip_set_default_password z password'
+         if r /= 0
+            then get_error z >>= E.throwIO
+            else withOpenArchive z action
+
+
+withOpenArchive :: Zip -> Archive a -> IO a
+withOpenArchive z action = do
       r <- fst `liftM` runStateT action z
       e <- c'zip_close z
       if e /= 0
         then get_error z >>= E.throwIO
         else return r
 
+
 -- | Get the number of entries in the archive.
-numFiles :: [FileFlag] -> Archive Integer
+numFiles :: [FileFlag]  -- ^ 'FileUNCHANGED' can be used to return
+                        -- the original unchanged number of entries.
+         -> Archive Integer
 numFiles flags  = do
   z <- getZip
   lift $ fromIntegral `liftM` c'zip_get_num_entries z (combine flags)
 
 -- | Get name of an entry in the archive by its index.
-fileName :: [FileFlag]  -- ^ 'FileUNCHANGED' flag can be used.
+fileName :: [FileFlag]  -- ^ 'FileUNCHANGED' flag can be used to
+                        -- return the original unchanged filename.
          -> Integer     -- ^ Position index of a file in the archive.
          -> Archive FilePath  -- ^ Name of the file in the archive.
 fileName flags i = do
@@ -169,7 +203,19 @@ fileName flags i = do
     doIf' (n /= nullPtr) z $ peekCString n
 
 -- | Locate an entry (get its index) in the archive by its name.
-nameLocate :: [FileFlag]  -- ^ Filename lookup mode.
+nameLocate :: [FileFlag]
+                -- ^ Filename lookup mode.
+                --   'FileNOCASE':     ignore case distinctions (only for ASCII).
+                --   'FileNODIR':      ignore directory part of the file name.
+                --   'FileENC_RAW':    compare against unmodified names as it is
+                --                     in the ZIP archive.
+                --   'FileENC_GUESS':  (default) guess encoding of the name in
+                --                     the ZIP archive and convert it to UTF-8,
+                --                     if necessary.
+                --   'FileENC_STRICT': follow the ZIP specification and expect
+                --                     CP-437 encoded names in the ZIP archive
+                --                     (except if they are explicitly marked as
+                --                     UTF-8). Convert it to UTF-8 before comparing.
            -> FilePath    -- ^ Name of the file in the archive.
            -> Archive (Maybe Integer)  -- ^ 'Just' position index if found.
 nameLocate flags name = do
@@ -225,7 +271,7 @@ fileStatIx flags i = do
        doIf' (r == 0) z $ toZipStat =<< peek stat
 
 -- | Delete file from the archive.
-deleteFile :: [FileFlag]  -- ^ Filename lookup mode.
+deleteFile :: [FileFlag]  -- ^ Filename lookup mode (see 'nameLocate').
            -> FilePath    -- ^ Filename.
            -> Archive ()
 deleteFile flags name = do
@@ -243,21 +289,28 @@ deleteFileIx i = do
      else lift $ get_error z >>= E.throwIO
 
 -- | Rename file in the archive.
-renameFile :: [FileFlag]  -- ^ Filename lookup mode.
+renameFile :: [FileFlag]  -- ^ Filename lookup mode (see 'nameLocate').
            -> FilePath    -- ^ Old name.
            -> FilePath    -- ^ New name.
            -> Archive ()
 renameFile flags oldname newname = do
   mbi <- nameLocate flags oldname
-  maybe (lift $ E.throwIO ErrNOENT) (\i -> renameFileIx i newname) mbi
+  maybe (lift $ E.throwIO ErrNOENT)
+            (\i -> renameFileIx i (UTF8.fromString newname) [FileENC_UTF_8])
+            mbi
 
 -- | Rename file (referenced by position index) in the archive.
-renameFileIx :: Integer  -- ^ Position index of a file in the archive.
-             -> FilePath -- ^ New name.
+renameFileIx :: Integer       -- ^ Position index of a file in the archive.
+             -> BS.ByteString -- ^ New name.
+             -> [FileFlag]    -- ^ Name encoding flags.
+                              -- 'FileENC_GUESS': guess encoding of the name (default).
+                              -- 'FileENC_UTF_8': interpret name as UTF-8.
+                              -- 'FileENC_CP437': interpret name as CP-437.
              -> Archive ()
-renameFileIx i newname = do
+renameFileIx i newname flags = do
   z <- getZip
-  r <- lift $ withCString newname $ c'zip_rename z (fromIntegral i)
+  r <- lift $ BS.useAsCString newname $ \s ->
+       c'zip_file_rename z (fromIntegral i) s (combine flags)
   if r == 0
      then return ()
      else lift $ get_error z >>= E.throwIO
@@ -266,10 +319,21 @@ renameFileIx i newname = do
 addFile :: FilePath   -- ^ Name of the file to create.
         -> ZipSource  -- ^ Source where file data is obtained from.
         -> Archive Int  -- ^ Position index of the new file.
-addFile name src = do
+addFile name src =
+  let utf8name = UTF8.fromString name
+  in  addFileWithFlags [FileENC_UTF_8] utf8name src
+
+addFileWithFlags
+    :: [FileFlag]   -- ^ Can be a combination of 'FileOVERWRITE' and/or one of
+                    -- filename encoding flags: 'FileENC_GUESS' (default),
+                    -- 'FileENC_UTF_8', 'FileENC_CP437'.
+    -> BS.ByteString   -- ^ Name of the file to create.
+    -> ZipSource       -- ^ Source where file data is obtained from.
+    -> Archive Int     -- ^ Position index of the new file.
+addFileWithFlags flags namebytes src = do
   z <- getZip
-  lift $ withCString name $ \name' -> do
-    i <- c'zip_add z name' src
+  lift $ BS.useAsCString namebytes $ \name' -> do
+    i <- c'zip_file_add z name' src (combine flags)
     if i < 0
        then c'zip_source_free src >> get_error z >>= E.throwIO
        else return $ fromIntegral i
@@ -277,15 +341,26 @@ addFile name src = do
 -- | Add a directory to the archive.
 addDirectory :: FilePath     -- ^ Directory's name in the archive.
              -> Archive Int  -- ^ Position index of the new directory entry.
-addDirectory name = do
+addDirectory name =
+  let utf8name = UTF8.fromString name
+  in  addDirectoryWithFlags [FileENC_UTF_8] utf8name
+
+-- | Add a directory to the archive.
+addDirectoryWithFlags
+    :: [FileFlag]        -- ^ Can be one of filename encoding flags:
+                         -- 'FileENC_GUESS (default), 'FileENC_UTF_8', 'FileENC_CP437'.
+    -> BS.ByteString     -- ^ Directory's name in the archive.
+    -> Archive Int       -- ^ Position index of the new directory entry.
+addDirectoryWithFlags flags name = do
   z <- getZip
-  r <- lift $ withCString name $ c'zip_add_dir z
+  r <- lift $ BS.useAsCString name $
+       \name'-> c'zip_dir_add z name' (combine flags)
   if r < 0
      then lift $ get_error z >>= E.throwIO
      else return (fromIntegral r)
 
 -- | Replace a file in the archive.
-replaceFile :: [FileFlag]  -- ^ Filename lookup mode.
+replaceFile :: [FileFlag]  -- ^ Filename lookup mode (see 'nameLocate').
             -> FilePath    -- ^ File to replace.
             -> ZipSource   -- ^ Source where the new file data is obtained from.
             -> Archive ()
@@ -294,6 +369,33 @@ replaceFile flags name src = do
   maybe (lift $ c'zip_source_free src >> E.throwIO ErrNOENT)
         (\i -> replaceFileIx i src >> return ()) mbi
 
+-- | Set compression method for a file in the archive.
+setFileCompression
+    :: [FileFlag]   -- ^ Filename lookup mode (see 'nameLocate').
+    -> FilePath     -- ^ Filename.
+    -> ZipCompMethod  -- ^ Compression method.
+                      -- As of libzip 0.11, the following methods are supported:
+                      -- 'CompDEFAULT', 'CompSTORE', 'CompDEFLATE'.
+    -> Archive ()
+setFileCompression flags name method = do
+  mbi <- nameLocate flags name
+  maybe (lift $ E.throwIO ErrNOENT) (\i -> setFileCompressionIx i method) mbi
+
+-- | Set compression method for a file in the archive.
+setFileCompressionIx
+    :: Integer   -- ^ Position index of a file in the archive.
+    -> ZipCompMethod  -- ^ Compression method.
+                      -- As of libzip 0.11, the following methods are supported:
+                      -- 'CompDEFAULT', 'CompSTORE', 'CompDEFLATE'.
+    -> Archive ()
+setFileCompressionIx i method = do
+  z <- getZip
+  lift $ do
+    r <- c'zip_set_file_compression z (fromIntegral i) (fromIntegral . fromEnum $ method) 0
+    if r /= 0
+       then get_error z >>= E.throwIO
+       else return ()
+
 -- | Replace a file in the archive (referenced by position index).
 replaceFileIx :: Integer   -- ^ Position index of a file in the archive.
               -> ZipSource -- ^ Source where the new file data is obtained from
@@ -301,7 +403,7 @@ replaceFileIx :: Integer   -- ^ Position index of a file in the archive.
 replaceFileIx i src = do
   z <- getZip
   lift $ do
-    r <- c'zip_replace z (fromIntegral i) src
+    r <- c'zip_file_replace z (fromIntegral i) src 0
     if r < 0
        then c'zip_source_free src >> get_error z >>= E.throwIO
        else return ()
@@ -412,7 +514,8 @@ runPureSource src pState pData len cmd
   | otherwise = return (-1)
 
 -- | Get zip archive comment.
-getComment :: [FileFlag]  -- ^ 'FileUNCHANGED' can be used.
+getComment :: [FileFlag]  -- ^ Can be a combination of 'FileUNCHANGED' and/or
+                          -- one of 'FileENC_GUESS' (default), 'FileENC_STRICT' (CP-437).
            -> Archive (Maybe String)
 getComment flags = do
   z <- getZip
@@ -422,14 +525,15 @@ getComment flags = do
          return (c,n)
   if  c == nullPtr
     then return Nothing
-    else lift $ peekCString c >>= return . Just . take (fromIntegral n)
+    else lift $ BS.packCStringLen (c, fromIntegral n) >>= return . Just . UTF8.toString
 
 -- | Set zip archive comment.
 setComment :: String   -- ^ Comment message.
            -> Archive ()
 setComment msg = do
   z <- getZip
-  r <- lift $ withCStringLen msg $ \(msg',i') ->
+  let utf8msg = UTF8.fromString msg
+  r <- lift $ BS.useAsCStringLen utf8msg $ \(msg',i') ->
        c'zip_set_archive_comment z msg' (fromIntegral i')
   if r < 0
      then lift $ get_error z >>= E.throwIO
@@ -445,50 +549,75 @@ removeComment = do
      else return ()
 
 -- | Get comment for a file in the archive.
-getFileComment :: [FileFlag]  -- ^ Filename lookup mode.
+getFileComment :: [FileFlag]  -- ^ Filename lookup mode (see 'nameLocate').
                -> FilePath    -- ^ Filename
                -> Archive (Maybe String)
 getFileComment flags name = do
   mbi <- nameLocate flags name
-  maybe (lift $ E.throwIO ErrNOENT) (getFileCommentIx flags) mbi
+  -- Backwards compatibility with LibZip < 0.11: FileUNCHANGED flag from
+  -- the filename lookup mode was used to get the original unchanged comment.
+  -- Please don't rely on this feature and use 'getFileCommentIx' instead.
+  let comment_flags = filter (== FileUNCHANGED) flags
+  maybe (lift $ E.throwIO ErrNOENT)
+            (\i -> do
+                 mbs <- getFileCommentIx comment_flags i
+                 -- 'FileENC_GUESS' is default => mbs is UTF-8 encoded
+                 return $ liftM UTF8.toString mbs
+            ) mbi
 
 -- | Get comment for a file in the archive (referenced by position index).
-getFileCommentIx :: [FileFlag]  -- ^ FileUNCHANGED can be used.
+getFileCommentIx :: [FileFlag]  -- ^ Comment lookup flags.
+                          --   'FileUNCHANGED':  return the original unchanged comment.
+                          --   'FileENC_RAW':    return the unmodified commment as it is.
+                          --   'FileENC_GUESS':  (default) guess the encoding of the comment
+                          --                     and convert it to UTF-8, if necessary.
+                          --   'FileENC_STRICT': follow the ZIP specification for file names
+                          --                     and extend it to file comments, expect
+                          --                     them to be encoded in CP-437. Convert it
+                          --                     to UTF-8.
                  -> Integer     -- ^ Position index of the file.
-                 -> Archive (Maybe String)
+                 -> Archive (Maybe BS.ByteString)
 getFileCommentIx flags i = do
   z <- getZip
   (c,n) <- lift $ alloca $ \lenp -> do
-           c <- c'zip_get_file_comment z (fromIntegral i) lenp (combine flags)
+           c <- c'zip_file_get_comment z (fromIntegral i) lenp (combine flags)
            n <- peek lenp
            return (c,n)
   if c == nullPtr
      then return Nothing
-     else lift $ peekCString c >>= return . Just . take (fromIntegral n)
+     else lift $ BS.packCStringLen (c,fromIntegral n) >>= return . Just
 
 -- | Set comment for a file in the archive.
-setFileComment :: [FileFlag]   -- ^ Name lookup mode.
+setFileComment :: [FileFlag]   -- ^ Filename lookup mode (see 'nameLocate').
                -> FilePath     -- ^ Filename.
                -> String       -- ^ New file comment.
                -> Archive ()
 setFileComment flags path comment = do
   mbi <- nameLocate flags path
-  maybe (lift $ E.throwIO ErrNOENT) (flip setFileCommentIx comment) mbi
+  let utf8comment = UTF8.fromString comment
+  let cflags = [FileENC_UTF_8]
+  maybe (lift $ E.throwIO ErrNOENT)
+            (\i -> setFileCommentIx i utf8comment cflags)
+            mbi
 
 -- | Set comment for a file in the archive (referenced by position index).
-setFileCommentIx :: Integer    -- ^ Position index of a file in the archive.
-                 -> String     -- ^ New file comment.
+setFileCommentIx :: Integer        -- ^ Position index of a file in the archive.
+                 -> BS.ByteString  -- ^ New file comment.
+                 -> [FileFlag]     -- ^ Comment encoding flags.
+                                   -- 'FileENC_GUESS': guess encoding of the comment (default).
+                                   -- 'FileENC_UTF_8': interpret comment as UTF-8.
+                                   -- 'FileENC_CP437': interpret comment as CP-437.
                  -> Archive ()
-setFileCommentIx i comment = do
+setFileCommentIx i comment cflags = do
   z <- getZip
-  r <- lift $ withCStringLen comment $ \(msg,len) ->
-       c'zip_set_file_comment z (fromIntegral i) msg (fromIntegral len)
+  r <- lift $ BS.useAsCStringLen comment $ \(msg,len) ->
+       c'zip_file_set_comment z (fromIntegral i) msg (fromIntegral len) (combine cflags)
   if r < 0
      then lift $ get_error z >>= E.throwIO
      else return ()
 
 -- | Remove comment for a file in the archive.
-removeFileComment :: [FileFlag]  -- ^ Filename lookup mode.
+removeFileComment :: [FileFlag]  -- ^ Filename lookup mode (see 'nameLocate').
                   -> FilePath    -- ^ Filename.
                   -> Archive ()
 removeFileComment flags path = do
@@ -499,14 +628,15 @@ removeFileComment flags path = do
 removeFileCommentIx :: Integer -- ^ Position index of a file in the archive.
                     -> Archive ()
 removeFileCommentIx i = do
+  let flags = 0   -- file name encoding flags (*_FL_*) are irrelevant
   z <- getZip
-  r <- lift $ c'zip_set_file_comment z (fromIntegral i) nullPtr 0
+  r <- lift $ c'zip_file_set_comment z (fromIntegral i) nullPtr 0 flags
   if r < 0
      then lift $ get_error z >>= E.throwIO
      else return ()
 
 -- | Undo changes to a file in the archive.
-unchangeFile :: [FileFlag]  -- ^ Filename lookup mode.
+unchangeFile :: [FileFlag]  -- ^ Filename lookup mode (see 'nameLocate').
              -> FilePath    -- ^ Filename.
              -> Archive ()
 unchangeFile flags name = do
