@@ -15,7 +15,7 @@ module Codec.Archive.LibZip.Types
     , ZipEncryptionMethod(..)
     , combine
     ) where
-    
+
 import Data.Bits ((.|.))
 import Data.Time (UTCTime)
 import Data.Time.Clock.POSIX (posixSecondsToUTCTime)
@@ -35,6 +35,7 @@ type Zip = Ptr C'zip
 type ZipFile = Ptr C'zip_file
 
 -- | Handler of data source for new files in the zip archive.
+-- Constructors: 'sourceBuffer', 'sourceFile', 'sourceZip', 'sourcePure'.
 type ZipSource = Ptr C'zip_source
 
 -- |  File statistics expressed in native Haskell types.
@@ -67,22 +68,25 @@ toZipStat s = do
     let flags =  toEnum . fromIntegral $ c'zip_stat'flags s
     return $ ZipStat valid name idx size comp_size mtime crc
              comp_meth enc_meth flags
-   
+
 
 -- | Flags for opening an archive.
 data OpenFlag
   = CreateFlag      -- ^ Create an archive if it does not exist.
   | ExclFlag        -- ^ Error if the archive already exists.
   | CheckConsFlag   -- ^ Check archive's consistency and error on failure.
+  | TruncateFlag    -- ^ If archive exists, ignore its current content.
   deriving (Show,Eq)
 
 instance Enum OpenFlag where
   fromEnum CheckConsFlag = c'ZIP_CHECKCONS
   fromEnum CreateFlag = c'ZIP_CREATE
   fromEnum ExclFlag = c'ZIP_EXCL
+  fromEnum TruncateFlag = c'ZIP_TRUNCATE
   toEnum x | x == c'ZIP_CHECKCONS = CheckConsFlag
   toEnum x | x == c'ZIP_CREATE = CreateFlag
   toEnum x | x == c'ZIP_EXCL = ExclFlag
+  toEnum x | x == c'ZIP_TRUNCATE = TruncateFlag
   toEnum _ = undefined
 
 -- | Flags for accessing files in the archive.
@@ -94,6 +98,14 @@ data FileFlag
   | FileUNCHANGED   -- ^ Read the original data, ignore changes.
   | FileRECOMPRESS  -- ^ Force recompression of data.
   | FileENCRYPTED   -- ^ Read encrypted data (implies FileCOMPRESSED).
+  | FileENC_GUESS   -- ^ Guess string encoding (default).
+  | FileENC_RAW     -- ^ Get unmodified string.
+  | FileENC_STRICT  -- ^ Follow specification strictly.
+  | FileLOCAL       -- ^ In local header.
+  | FileCENTRAL     -- ^ In central directory.
+  | FileENC_UTF_8   -- ^ String is UTF-8 encoded.
+  | FileENC_CP437   -- ^ String is CP437 encoded.
+  | FileOVERWRITE   -- ^ When adding files: if file name exists, overwrite.
   deriving (Show,Eq)
 
 instance Enum FileFlag where
@@ -103,12 +115,28 @@ instance Enum FileFlag where
   fromEnum FileRECOMPRESS = c'ZIP_FL_RECOMPRESS
   fromEnum FileUNCHANGED = c'ZIP_FL_UNCHANGED
   fromEnum FileENCRYPTED = c'ZIP_FL_ENCRYPTED
+  fromEnum FileENC_GUESS = c'ZIP_FL_ENC_GUESS
+  fromEnum FileENC_RAW = c'ZIP_FL_ENC_RAW
+  fromEnum FileENC_STRICT = c'ZIP_FL_ENC_STRICT
+  fromEnum FileLOCAL = c'ZIP_FL_LOCAL
+  fromEnum FileCENTRAL = c'ZIP_FL_CENTRAL
+  fromEnum FileENC_UTF_8 = c'ZIP_FL_ENC_UTF_8
+  fromEnum FileENC_CP437 = c'ZIP_FL_ENC_CP437
+  fromEnum FileOVERWRITE = c'ZIP_FL_OVERWRITE
   toEnum x | x == c'ZIP_FL_COMPRESSED = FileCOMPRESSED
   toEnum x | x == c'ZIP_FL_NOCASE = FileNOCASE
   toEnum x | x == c'ZIP_FL_NODIR = FileNODIR
   toEnum x | x == c'ZIP_FL_RECOMPRESS = FileRECOMPRESS
   toEnum x | x == c'ZIP_FL_UNCHANGED = FileUNCHANGED
   toEnum x | x == c'ZIP_FL_ENCRYPTED = FileENCRYPTED
+  toEnum x | x == c'ZIP_FL_ENC_GUESS = FileENC_GUESS
+  toEnum x | x == c'ZIP_FL_ENC_RAW = FileENC_RAW
+  toEnum x | x == c'ZIP_FL_ENC_STRICT = FileENC_STRICT
+  toEnum x | x == c'ZIP_FL_LOCAL = FileLOCAL
+  toEnum x | x == c'ZIP_FL_CENTRAL = FileCENTRAL
+  toEnum x | x == c'ZIP_FL_ENC_UTF_8 = FileENC_UTF_8
+  toEnum x | x == c'ZIP_FL_ENC_CP437 = FileENC_CP437
+  toEnum x | x == c'ZIP_FL_OVERWRITE = FileOVERWRITE
   toEnum _ = undefined
 
 
@@ -127,11 +155,16 @@ instance Enum ArchiveFlag where
 
 
 -- | @libzip@ flags for compression and encryption sources
-data CodecFlag = CodecENCODE deriving (Show, Eq)
+data CodecFlag
+    = CodecENCODE
+    | CodecDECODE
+    deriving (Show, Eq)
 
 instance Enum CodecFlag where
     fromEnum CodecENCODE = c'ZIP_CODEC_ENCODE
+    fromEnum CodecDECODE = c'ZIP_CODEC_DECODE
     toEnum x | x == c'ZIP_CODEC_ENCODE = CodecENCODE
+    toEnum x | x == c'ZIP_CODEC_DECODE = CodecDECODE
     toEnum _ = undefined
 
 
@@ -254,7 +287,7 @@ instance Show ZipError where
   show ErrREMOVE         =  "Can't remove file"
   show ErrDELETED        =  "Entry has been deleted"
   show ErrENCRNOTSUPP    =  "Encryption method not supported"
-  show ErrRDONLY         =  "Read-only archive" 
+  show ErrRDONLY         =  "Read-only archive"
   show ErrNOPASSWD       =  "No password provided"
   show ErrWRONGPASSWD    =  "Wrong password provided"
 
@@ -334,4 +367,3 @@ instance Enum ZipEncryptionMethod where
 
 combine :: (Enum a, Num b) => [a] -> b
 combine fs = fromIntegral . foldr (.|.) 0 $ map fromEnum fs
-
