@@ -1,5 +1,5 @@
 module Tests.Common
-    ( testzip, testfiles, lastfile, lastfilesize
+    ( testzip, encryptedzip, testfiles, lastfile, lastfilesize
     , world_txt
     , toUpper, toLower, map2
     ) where
@@ -7,6 +7,7 @@ module Tests.Common
 import Data.Char (toUpper, toLower)
 
 testzip = "Tests/test.zip"
+encryptedzip = "Tests/encrypted.zip"
 testfiles = [ "hello/", "hello/world.txt" ]
 lastfile = last testfiles
 lastfilesize = 71 :: Integer
