@@ -1,12 +1,10 @@
-import Tests.LegacyTests (legacyTests)
 import Tests.MonadicTests (monadicTests)
 
 import System.Exit
 import Test.HUnit
 
 allTests = TestList
-  [ "Legacy API" ~: legacyTests
-  , "Monadic API" ~: monadicTests
+  [ "Monadic API" ~: monadicTests
   ]
 
 main = do
